#!/usr/bin/env cwl-runner
cwlVersion: v1.2
class: Workflow
requirements:
  StepInputExpressionRequirement: {}
  InlineJavascriptRequirement: {}
  MultipleInputFeatureRequirement: {}
  ScatterFeatureRequirement: {}
  SubworkflowFeatureRequirement: {}

label: Long Read Quality Control and Filtering
doc: |
  **Workflow for long read quality control and contamination filtering.**
  - NanoPlot before and after filtering (read quality control)
  - Filtlong filter on quality and length
  - Kraken2 taxonomic read classification (before and after filtering)
  - Minimap2 read filtering based on given references

  Other UNLOCK workflows on WorkflowHub: https://workflowhub.eu/projects/16/workflows?view=default<br><br>

  **All tool CWL files and other workflows can be found here:**<br>
    Tools: https://gitlab.com/m-unlock/cwl/-/tree/master/cwl <br>
    Workflows: https://gitlab.com/m-unlock/cwl/-/tree/master/cwl/workflows<br>

  **How to setup and use an UNLOCK workflow:**<br>
  https://m-unlock.gitlab.io/docs/setup/setup.html<br>

outputs:
  filtlong_log:
    type: File
    label: Filtlong log
    doc: Log file from filtlong longread filtering
    outputSource: filtlong/log  
  reference_filter_longreads_log:
    type: File?
    label: Reference filter log
    doc: Log file from minimap2 from reference filter.
    outputSource: reference_filter_longreads/log

  nanoplot_unfiltered_folder:
    type: Directory?
    label: NanoPlot unfiltered
    doc: Folder with quality plots from Nanoplot before filtering
    outputSource: nanoplot_unfiltered_files_to_folder/results
  nanoplot_filtered_folder:
    type: Directory
    label: NanoPlot filtered
    doc: Folder with quality plots from Nanoplot after filtering
    outputSource: nanoplot_filtered_files_to_folder/results
  kraken2_folder:
    type: Directory?
    label: Kraken2 folder
    doc: Folder with Kraken2 output files
    outputSource: kraken2_files_to_folder/results
  filtered_reads:
    type: File
    label: Filtered long reads
    doc: Filtered long reads
    outputSource:
      - reference_filter_longreads/fastq
      - filtlong/output_reads
    pickValue: first_non_null

inputs:
  identifier:
    type: string
    doc: Identifier for this dataset used in this workflow
    label: identifier used
  longreads:
    type: File[]
    doc: Long read sequence file locally fastq format
    label: Long reads
  readtype:
    type: string
    doc: Type of read i.e. PacBio or Nanopore. Used for naming output files.
    label: Read type
  fastq_rich:
    type: boolean
    doc: | 
          Input fastq is generated by albacore, MinKNOW or guppy  with additional information concerning channel and time. 
          Used to creating more informative quality plots (default false)
    label: Fastq rich (ONT)
    default: false

  skip_qc_unfiltered:
    type: boolean?
    doc: Skip NanoPlot analyses of unfiltered input data (default false)
    label: Skip NanoPlot before
    default: false
  skip_qc_filtered:
    type: boolean?
    doc: Skip NanoPlot analyses of filter input data (default false)
    label: Skip NanoPlot after
    default: false
    

# Defaults taken from nf-core MAG
  minimum_length:
    type: int?
    doc: Minimum read length threshold (default 1000)
    label: Minimum read length
    default: 1000
  keep_percent:
    type: float?
    doc: Maximum read length threshold (default 90)
    label: Maximum read length threshold
    default: 90
  length_weight:
    type: float?
    doc: Weight given to the length score (default 10)
    label: Length weigth
    default: 10

  filter_references:
    type: File[]?
    doc: Contamination references fasta file for contamination filtering. Gzipped or not (Not mixed)
    label: Contamination reference file

  keep_reference_mapped_reads:
    type: boolean?
    doc: Keep with reads mapped to the given reference (default false)
    label: Keep mapped reads
    default: false
  prepare_reference:
    type: boolean?
    doc: Prepare reference with unique headers (default true)
    label: Prepare references
    default: true

  kraken2_database:
    type: Directory[]?
    label: Kraken2 database
    doc: Kraken2 database location, multiple databases is possible
    default: []
  kraken2_confidence:
    type: float?
    label: Kraken2 confidence threshold
    doc: Confidence score threshold (default 0.0) must be between [0, 1]
  kraken2_standard_report:
    type: boolean
    label: Kraken2 standard report
    doc: Also output Kraken2 standard report with per read classification. These can be large. (default false)
    default: false
    
  threads:
    type: int?
    doc: Number of threads to use for computational processes
    label: Number of threads
    default: 2
  memory:
    type: int?
    doc: Maximum memory usage in megabytes
    label: Maximum memory in MB
    default: 4000

  step:
    type: int?
    label: CWL base step number
    doc: Step number for order of steps
    default: 1

  destination:
    type: string?
    label: Output Destination
    doc: Optional Output destination used for cwl-prov reporting.

steps:
#############################################
#### merging of FASTQ files to only one
  merge_longreads_fastq:
    label: Merge fastq files
    when: $(inputs.longreads.length > 1)
    run: ../tools/bash/concatenate.cwl
    in:
      longreads: longreads
      identifier: identifier
      readtype: readtype

      infiles: longreads
      outname:
        valueFrom: $(inputs.identifier)_$(inputs.readtype)_merged_raw.fastq.gz
    out: [output]
#############################################
#### Pick first file of longreads when only 1 file is given
  longreads_array_to_file:
    label: Array to file
    doc: Pick first file of longreads when only 1 file is given
    when: $(inputs.longreads.length === 1)
    run: ../tools/expressions/array_to_file_tool.cwl
    in:
      longreads: longreads

      files: longreads
    out: [file]
#############################################
#### NanoPlot Before
  nanoplot_longreads_unfiltered:
    label: NanoPlot unfiltered
    doc: NanoPlot Quality assessment and report of reads before filtering
    run: ../tools/nanoplot/nanoplot.cwl
    when: $(inputs.skip_nanoplot_unfiltered == false)
    in:
      skip_nanoplot_unfiltered: skip_qc_unfiltered
      identifier: identifier
      readtype: readtype
      fastq_rich: fastq_rich

      plot_title:
        valueFrom: $(inputs.identifier)_$(inputs.readtype)_unfiltered
      file_prefix:
        valueFrom: $(inputs.identifier)_$(inputs.readtype)_unfiltered_

      array_to_file: longreads_array_to_file/file
      merged_files: merge_longreads_fastq/output
      fastq_files:
        valueFrom:
          ${
            if (inputs.array_to_file !== null && inputs.fastq_rich == false) {
              var fastq = [inputs.array_to_file];
            } else if (inputs.merged_files !== null && inputs.fastq_rich == false) {
              var fastq = [inputs.merged_files];
            } else {
              var fastq = null
            }
            return fastq;
          }
      rich_fastq_files:
        valueFrom:
          ${
            if (inputs.array_to_file !== null && inputs.fastq_rich) {
              var fastq = [inputs.array_to_file];
            } else if (inputs.merged_files !== null && inputs.fastq_rich) {
              var fastq = [inputs.merged_files];
            } else {
              var fastq = null
            }
            return fastq;
          }
      store_pickle:
        default: true
      store_raw:
        default: true
      threads: threads
    out: [log, WeightedHistogramReadlength, WeightedLogTransformed_HistogramReadlength, Non_weightedHistogramReadlength, Non_weightedLogTransformed_HistogramReadlength, LengthvsQualityScatterPlot, Yield_By_Length, report, logfile, nanostats, pickle, raw_data, CumulativeYieldPlot_Gigabases, CumulativeYieldPlot_NumberOfReads, NumberOfReads_Over_Time, ActivePores_Over_Time, TimeLengthViolinPlot, TimeQualityViolinPlot]
#############################################
#### Filtlong
  filtlong:
    label: Filtlong
    doc: Filter longreads on quality and length
    run: ../tools/filtlong/filtlong.cwl
    in:
      identifier: identifier
      readtype: readtype

      output_filename: 
        valueFrom: $(inputs.identifier)_$(inputs.readtype)_filtered
      long_reads:
        source:
        - merge_longreads_fastq/output
        - longreads_array_to_file/file
        pickValue: first_non_null
      keep_percent: keep_percent
      minimum_length: minimum_length
      length_weight: length_weight
    out: [output_reads, log]
#############################################
#### Preparation of reference files
  prepare_fasta_db:
    label: Prepare references
    doc: Prepare references to a single fasta file and unique headers
    when: $(inputs.fasta_input !== null && inputs.fasta_input.length !== 0)
    run: workflow_prepare_fasta_db.cwl
    in:
      make_headers_unique: prepare_reference
      fasta_input: filter_references
      output_name: identifier
    out: [fasta_db]
#############################################
#### Reference filter
  reference_filter_longreads:
    label: Reference mapping
    doc: Removal of contaminated reads using minimap2 mapping
    when: $(inputs.filter_references !== null && inputs.filter_references.length !== 0)
    run: ../tools/minimap2/minimap2_to_fastq.cwl
    in:
      readtype: readtype
      filter_references: filter_references

      reference: prepare_fasta_db/fasta_db
      reads: filtlong/output_reads
      identifier:
        source: identifier
        valueFrom: $(self+"_"+inputs.readtype)
      threads: threads
      output_mapped: keep_reference_mapped_reads
      preset:
        default: "map-ont"
    out: [fastq, log]
#############################################
#### NanoPlot After
  nanoplot_longreads_filtered:
    label: NanoPlot before
    doc: NanoPlot Quality assessment and report of reads after filtering
    run: ../tools/nanoplot/nanoplot.cwl
    when: $(inputs.skip_nanoplot_filtered == false)
    in:
      skip_nanoplot_filtered: skip_qc_filtered
      identifier: identifier
      readtype: readtype

      plot_title: 
        valueFrom: $(inputs.identifier)_$(inputs.readtype)_filtered
      file_prefix:
        valueFrom: $(inputs.identifier)_$(inputs.readtype)_filtered_

      reffilt_reads: reference_filter_longreads/fastq
      filtlong_reads: filtlong/output_reads

      # Cannot use fastq_rich, minimap2 removes the extra info
      fastq_files:
        valueFrom:
          ${
            if (inputs.reffilt_reads !== null) {
              var fastq = [inputs.reffilt_reads];
            } else if (inputs.filtlong_reads !== null) {
              var fastq = [inputs.filtlong_reads];
            } else {
              var fastq = null
            }
            return fastq;
          }

      store_pickle:
        default: true
      store_raw:
        default: true
      threads: threads
    out: [log, WeightedHistogramReadlength, WeightedLogTransformed_HistogramReadlength, Non_weightedHistogramReadlength, Non_weightedLogTransformed_HistogramReadlength, LengthvsQualityScatterPlot, Yield_By_Length, report, logfile, nanostats, pickle, raw_data, CumulativeYieldPlot_Gigabases, CumulativeYieldPlot_NumberOfReads, NumberOfReads_Over_Time, ActivePores_Over_Time, TimeLengthViolinPlot, TimeQualityViolinPlot]

#############################################
#### Kraken2 BEFORE and AFTER filter
  longreads_kraken2_before:
    label: Kraken2 unfiltered
    doc: Taxonomic classification of FASTQ reads before filtering
    when: $(inputs.database !== null && inputs.database.length !== 0)
    run: ../tools/kraken2/kraken2.cwl
    scatter: database
    in:
      readtype: readtype

      identifier:
        source: identifier
        valueFrom: $(self+"_"+inputs.readtype)_unfiltered
      threads: threads
      database: kraken2_database
      confidence: kraken2_confidence
      nanopore_reads:
        source:
        - merge_longreads_fastq/output
        - longreads_array_to_file/file
        pickValue: first_non_null
    out: [sample_report, standard_report]

  longreads_kraken2_after:
    label: Kraken2 filtered
    doc: Taxonomic classification of FASTQ reads after filtering
    when: $(inputs.database !== null && inputs.database.length !== 0)
    run: ../tools/kraken2/kraken2.cwl
    scatter: database
    in:
      readtype: readtype

      identifier:
        source: identifier
        valueFrom: $(self+"_"+inputs.readtype)_filtered
      threads: threads
      database: kraken2_database
      confidence: kraken2_confidence
      nanopore_reads:
        source:
        - reference_filter_longreads/fastq
        - filtlong/output_reads
        pickValue: first_non_null
    out: [sample_report, standard_report]

  longreads_kraken2_compress:
    label: Compress kraken2
    doc: Compress large kraken2 report file
    when: $(inputs.kraken2_database !== null && inputs.kraken2_database.length !== 0  && inputs.kraken2_standard_report)
    run: ../tools/bash/pigz.cwl
    scatter: inputfile
    in:
      kraken2_database: kraken2_database
      kraken2_standard_report: kraken2_standard_report

      inputfile:
        source: [longreads_kraken2_before/standard_report, longreads_kraken2_after/standard_report]
        linkMerge: merge_flattened
        pickValue: all_non_null
      threads: threads
    out: [outfile]

  longreads_kraken2_krona:
    label: Krona Kraken2
    doc: Visualization of kraken2 with Krona
    when: $(inputs.kraken2_database !== null && inputs.kraken2_database.length !== 0)
    run: ../tools/krona/krona.cwl
    scatter: kraken
    in:
      kraken2_database: kraken2_database

      kraken:
        source: [longreads_kraken2_before/sample_report, longreads_kraken2_after/sample_report]
        linkMerge: merge_flattened
        pickValue: all_non_null
    out: [krona_html]

#############################################
#### Move files to a specific folder
  nanoplot_unfiltered_files_to_folder:
    label: Nanoplot unfiltered folder
    doc: Nanoplot plots and files to single folder
    when: $(inputs.skip_nanoplot_unfiltered == false)
    in:
      skip_nanoplot_unfiltered: skip_qc_unfiltered
      readtype: readtype

      files:
        source: [nanoplot_longreads_unfiltered/log, nanoplot_longreads_unfiltered/WeightedHistogramReadlength, nanoplot_longreads_unfiltered/WeightedLogTransformed_HistogramReadlength, nanoplot_longreads_unfiltered/Non_weightedHistogramReadlength, nanoplot_longreads_unfiltered/Non_weightedLogTransformed_HistogramReadlength, nanoplot_longreads_unfiltered/LengthvsQualityScatterPlot, nanoplot_longreads_unfiltered/Yield_By_Length, nanoplot_longreads_unfiltered/report, nanoplot_longreads_unfiltered/logfile, nanoplot_longreads_unfiltered/nanostats, nanoplot_longreads_unfiltered/pickle, nanoplot_longreads_unfiltered/raw_data, nanoplot_longreads_unfiltered/CumulativeYieldPlot_Gigabases, nanoplot_longreads_unfiltered/CumulativeYieldPlot_NumberOfReads, nanoplot_longreads_unfiltered/NumberOfReads_Over_Time, nanoplot_longreads_unfiltered/ActivePores_Over_Time, nanoplot_longreads_unfiltered/TimeLengthViolinPlot, nanoplot_longreads_unfiltered/TimeQualityViolinPlot]
        linkMerge: merge_flattened
        pickValue: all_non_null
      destination:
        source: readtype
        valueFrom: $(self+"_quality_reports_unfiltered")
    run: ../tools/expressions/files_to_folder_tool.cwl
    out:
      [results]
  nanoplot_filtered_files_to_folder:
    label: Nanoplot filtered folder
    doc: Nanoplot plots and files to single folder
    when: $(inputs.skip_nanoplot_filtered == false)
    in:
      skip_nanoplot_filtered: skip_qc_filtered
      readtype: readtype

      files:
        source: [nanoplot_longreads_filtered/log, nanoplot_longreads_filtered/WeightedHistogramReadlength, nanoplot_longreads_filtered/WeightedLogTransformed_HistogramReadlength, nanoplot_longreads_filtered/Non_weightedHistogramReadlength, nanoplot_longreads_filtered/Non_weightedLogTransformed_HistogramReadlength, nanoplot_longreads_filtered/LengthvsQualityScatterPlot, nanoplot_longreads_filtered/Yield_By_Length, nanoplot_longreads_filtered/report, nanoplot_longreads_filtered/logfile, nanoplot_longreads_filtered/nanostats, nanoplot_longreads_filtered/pickle, nanoplot_longreads_filtered/raw_data, nanoplot_longreads_filtered/CumulativeYieldPlot_Gigabases, nanoplot_longreads_filtered/CumulativeYieldPlot_NumberOfReads, nanoplot_longreads_filtered/NumberOfReads_Over_Time, nanoplot_longreads_filtered/ActivePores_Over_Time, nanoplot_longreads_filtered/TimeLengthViolinPlot, nanoplot_longreads_filtered/TimeQualityViolinPlot]
        linkMerge: merge_flattened
        pickValue: all_non_null
      destination:
        source: readtype
        valueFrom: $(self+"_quality_reports_filtered")
    run: ../tools/expressions/files_to_folder_tool.cwl
    out:
      [results]

  kraken2_files_to_folder:
    label: Kraken2 folder
    doc: Kraken2 files to single folder
    when: $(inputs.kraken2_database !== null && inputs.kraken2_database.length !== 0)
    in:
      kraken2_database: kraken2_database
      readtype: readtype

      files:
        source: [longreads_kraken2_before/sample_report, longreads_kraken2_after/sample_report, longreads_kraken2_krona/krona_html, longreads_kraken2_compress/outfile]
        linkMerge: merge_flattened
        pickValue: all_non_null
      destination:
        source: readtype
        valueFrom: $("Kraken2_"+self)
    run: ../tools/expressions/files_to_folder_tool.cwl
    out:
      [results]

s:author:
  - class: s:Person
    s:identifier: https://orcid.org/0000-0001-8172-8981
    s:email: mailto:jasper.koehorst@wur.nl
    s:name: Jasper Koehorst
  - class: s:Person
    s:identifier: https://orcid.org/0000-0001-9524-5964
    s:email: mailto:bart.nijsse@wur.nl
    s:name: Bart Nijsse

s:citation: https://m-unlock.nl
s:codeRepository: https://gitlab.com/m-unlock/cwl
s:dateCreated: "2020-00-00"
s:dateModified: "2023-04-03"
s:license: https://spdx.org/licenses/Apache-2.0
s:copyrightHolder: "UNLOCK - Unlocking Microbial Potential"


$namespaces:
  s: https://schema.org/
